import copy
import os
import re

# NOTE: we import direct from searchkit rather than hotsos.core.search to
#       avoid circular dependency issues.
from searchkit import (
    FileSearcher,
    SearchDef,
    SequenceSearchDef,
)
from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers.cli import CLIHelper, CLIHelperFile
from hotsos.core.host_helpers.common import HostHelpersBase
from hotsos.core.log import log
from hotsos.core.utils import mktemp_dump

# compatible with ip addr and ip link
# this one is name and state
IP_IFACE_NAME = r"^\d+:\s+(\S+):\s+.+mtu\s+(\d+).+state\s+(\S+)"
IP_IFACE_NAME_TEMPLATE = r"^\d+:\s+({}):\s+.+"
IP_IFACE_V4_ADDR = r".+(inet) ([\d\.]+)/(\d+) (?:brd \S+ )?scope global (\S+)"
IP_IFACE_V6_ADDR = r".+(inet6) (\S+)/(\d+) scope global (\S+)"
IP_IFACE_HW_ADDR = r".+(link/ether) (\S+) brd .+"
IP_IFACE_HW_ADDR_TEMPLATE = r".+(link/ether) {} brd .+"
IP_IFACE_VXLAN_INFO = r"\s+(vxlan) id (\d+) local (\S+) dev (\S+) .+"
IP_EOF = r"^$"


class NetworkPort(HostHelpersBase):

    def __init__(self, name, addresses, hwaddr, state, encap_info,
                 mtu, namespace=None):
        self.name = name
        self.addresses = addresses
        self.hwaddr = hwaddr
        self.state = state
        self.encap_info = encap_info
        self.mtu = mtu
        self.namespace = namespace
        super().__init__()

    @property
    def cache_root(self):
        """
        Cache this information at the plugin level rather than globally.
        """
        return HotSOSConfig.plugin_tmp_dir

    @property
    def cache_type(self):
        return 'network_ports'

    @property
    def cache_name(self):
        if self.namespace:
            return "ns-{}-port-{}".format(self.namespace, self.name)

        return "port-{}".format(self.name)

    def cache_load(self):
        contents = self.cache.get('stats')  # pylint: disable=E1111
        if not contents:
            log.debug("network port %s not found in cache", self.name)
            return

        log.debug("loading network port %s from cache", self.name)
        return contents

    def cache_save(self, data):
        log.debug("saving network port %s to cache", self.name)
        self.cache.set('stats', data)

    def to_dict(self):
        info = {'addresses': copy.deepcopy(self.addresses),
                'hwaddr': self.hwaddr,
                'mtu': self.mtu,
                'state': self.state,
                'speed': self.speed}
        if self.namespace:
            info['namespace'] = self.namespace

        return {self.name: info}

    @property
    def speed(self):
        # need to strip @* since sosreport does that too
        name = self.name.partition('@')[0]
        out = CLIHelper().ethtool(interface=name)
        if out:
            for line in out:
                ret = re.match(r'\s*Speed:\s+(\S+)', line)
                if ret:
                    return ret.group(1)

        return 'unknown'

    @property
    def stats(self):
        """ Get ip link info for the interface. """
        counters = self.cache_load() or {}
        if counters:
            return counters

        s = FileSearcher()
        seqdef = SequenceSearchDef(
                    # match start of interface
                    start=SearchDef(IP_IFACE_NAME_TEMPLATE.format(self.name)),
                    # match body of interface
                    body=SearchDef(r".+"),
                    # match next interface or EOF
                    end=SearchDef([IP_IFACE_NAME, IP_EOF]),
                    tag="ifaces")
        stats_raw = []
        with CLIHelperFile() as cli:
            s.add(seqdef, path=cli.ip_link())
            results = s.run()
            for section in results.find_sequence_sections(seqdef).values():
                for result in section:
                    if result.tag == seqdef.body_tag:
                        stats_raw.append(result.get(0))

        if not stats_raw:
            return {}

        # NOTE: we only expect one match
        for i, line in enumerate(stats_raw):
            ret = re.compile(r"\s+([RT]X):\s+.+").findall(line)
            if ret:
                rxtx = ret[0].lower()
                ret = re.compile(r"\s*([a-z]+)\s*").findall(line)
                if ret:
                    for j, column in enumerate(ret):
                        value = int(stats_raw[i + 1].split()[j])
                        if column in ['packets', 'dropped', 'errors',
                                      'overrun']:
                            if rxtx not in counters:
                                counters[rxtx] = {}

                            counters[rxtx][column] = value

        if counters:
            self.cache_save(counters)
            return counters

        return {}


class HostNetworkingHelper(HostHelpersBase):

    def __init__(self):
        super().__init__()
        self._host_interfaces = None
        self._host_ns_interfaces = None
        self.cli = CLIHelper()

    @property
    def cache_type(self):
        return 'networks'

    @property
    def cache_name(self):
        # this is a global cache i.e. shared by all plugins
        return 'network-helper'

    def cache_load(self, all_namespaces=False, namespace=None):
        """
        @param all_namespaces: Set to True to if we are retrieving network info
                               for all namespaces.
        @param namespace: Name of specific namespace we are retrieving from.
                          Takes precedence over all_namespaces.
        """
        log.debug("loading network helper info from cache (all_namespaces=%s, "
                  "namespace=%s)", all_namespaces, namespace)
        if namespace is not None:
            contents = self.cache.get('ns-{}-interfaces'.format(namespace))  # noqa, pylint: disable=E1111
        elif all_namespaces:
            contents = self.cache.get('ns-interfaces')  # pylint: disable=E1111
        else:
            contents = self.cache.get('interfaces')  # pylint: disable=E1111

        if contents:
            return contents

        log.debug("network helper info not available in cache "
                  "(all_namespaces=%s, namespace=%s)", all_namespaces,
                  namespace)

    def cache_save(self, data, all_namespaces=False, namespace=None):
        log.debug("saving network helper info to cache (all_namespaces=%s, "
                  "namespaces=%s)", all_namespaces, namespace)
        if namespace is not None:
            self.cache.set('ns-{}-interfaces'.format(namespace), data)
        elif all_namespaces:
            self.cache.set('ns-interfaces', data)
        else:
            self.cache.set('interfaces', data)

    def _extract_iface_info(self, seqdef, section, search_obj):
        addrs = []
        encap_info = None
        hwaddr = None
        name = None
        mtu = None
        state = None
        for result in section:
            # infer ns name from filename prefix
            ns_name = None
            source = search_obj.resolve_source_id(result.source_id)
            ret = re.match(r'^__ns_start__(\S+)__ns__end__',
                           os.path.basename(source))
            if ret:
                ns_name = ret.group(1)

            if result.tag == seqdef.start_tag:
                name = result.get(1)
                mtu = int(result.get(2))
                state = result.get(3)
            elif result.tag == seqdef.body_tag:
                if result.get(1) in ['inet', 'inet6']:
                    addrs.append(result.get(2))
                elif result.get(1) in ['vxlan']:
                    encap_info = {result.get(1): {
                                      'id': result.get(2),
                                      'local_ip': result.get(3),
                                      'dev': result.get(4)}}
                else:
                    hwaddr = result.get(2)

        return {'name': name, 'addresses': addrs, 'hwaddr': hwaddr, 'mtu': mtu,
                'state': state, 'encap_info': encap_info, 'namespace': ns_name}

    @property
    def _ip_addr_show_iface_sequence_def(self):
        return SequenceSearchDef(start=SearchDef(IP_IFACE_NAME),
                                 body=SearchDef([IP_IFACE_V4_ADDR,
                                                 IP_IFACE_V6_ADDR,
                                                 IP_IFACE_HW_ADDR,
                                                 IP_IFACE_VXLAN_INFO]),
                                 tag='ip_addr_show')

    def get_ns_interfaces(self, namespace):
        """ Get all network ports for the given namespace.

        @param namespace: name of namespace
        @return: generator of NetworkPort objects
        """
        interfaces_raw = self.cache_load(namespace=namespace) or []
        if not interfaces_raw:
            seq = self._ip_addr_show_iface_sequence_def
            search_obj = FileSearcher()
            ip_addr = self.cli.ns_ip_addr(namespace=namespace)
            prefix = "__ns_start__{}__ns__end__".format(namespace)
            path = mktemp_dump(''.join(ip_addr), prefix=prefix)
            try:
                search_obj.add(seq, path)
                r = search_obj.run()
                sections = r.find_sequence_sections(seq, path).values()
                for section in sections:
                    interfaces_raw.append(self._extract_iface_info(seq,
                                                                   section,
                                                                   search_obj))
            finally:
                os.unlink(path)

            self.cache_save(interfaces_raw, namespace=namespace)

        for iface in interfaces_raw:
            yield NetworkPort(**iface)

    def _get_interfaces(self, all_namespaces=False):
        """
        Get all interfaces in ip address show.

        @param all_namespaces: if set to True will get interfaces from all
        namespaces on the host.
        @return: list of NetworkPort objects for each interface found.
        """
        interfaces_raw = self.cache_load(all_namespaces=all_namespaces) or []
        if not interfaces_raw:
            seq = self._ip_addr_show_iface_sequence_def
            search_obj = FileSearcher()
            if all_namespaces:
                for ns in self.cli.ip_netns():
                    ns_name = ns.partition(" ")[0]
                    ip_addr = self.cli.ns_ip_addr(namespace=ns_name)
                    prefix = "__ns_start__{}__ns__end__".format(ns_name)
                    path = mktemp_dump(''.join(ip_addr), prefix=prefix)
                    search_obj.add(seq, path)
            else:
                path = mktemp_dump(''.join(self.cli.ip_addr()))
                search_obj.add(seq, path)

            if not search_obj.files:
                log.debug("no network info found (all_namespaces=%s)",
                          all_namespaces)
                return []

            r = search_obj.run()
            for path in search_obj.files:
                try:
                    sections = r.find_sequence_sections(seq, path).values()
                    for section in sections:
                        interfaces_raw.append(self._extract_iface_info(
                                                                seq,
                                                                section,
                                                                search_obj))
                finally:
                    os.unlink(path)

            self.cache_save(interfaces_raw, all_namespaces=all_namespaces)

        for iface in interfaces_raw:
            yield NetworkPort(**iface)

    @property
    def host_interfaces(self):
        if self._host_interfaces is not None:
            return self._host_interfaces

        self._host_interfaces = list(self._get_interfaces())
        return self._host_interfaces

    @property
    def host_ns_interfaces(self):
        if self._host_ns_interfaces is not None:
            return self._host_ns_interfaces

        self._host_ns_interfaces = list(
                                     self._get_interfaces(all_namespaces=True))
        return self._host_ns_interfaces

    @property
    def host_interfaces_all(self):
        return self.host_interfaces + self.host_ns_interfaces

    def get_interface_with_hwaddr(self, hwaddr):
        """ Returns first found. """
        for iface in self.host_interfaces_all:
            if iface.hwaddr == hwaddr:
                return iface

    def get_interface_with_addr(self, addr):
        for iface in self.host_interfaces_all:
            for _addr in iface.addresses:
                if _addr.startswith(addr):
                    return iface

    def get_interface_with_name(self, name):
        for iface in self.host_interfaces_all:
            if iface.name == name:
                return iface

    def host_interface_exists(self, name, check_namespaces=True):
        names = [_iface.name for _iface in self.host_interfaces]
        if name in names:
            return True

        if not check_namespaces:
            return False

        names = [_iface.name for _iface in self.host_ns_interfaces]
        if name in names:
            return True

        return False



from hotsos.core.plugins.lxd import LXD, LXDChecksBase


class LXDSummary(LXDChecksBase):
    summary_part_index = 0

    def __0_summary_services(self):
        if self.systemd.services:
            return self.systemd.summary

    def __1_summary_snaps(self):
        if self.snaps:
            return self.snaps.all_formatted

    def __2_summary_dpkg(self):
        if self.apt:
            return self.apt.all_formatted

    def __3_summary_instances(self):
        instances = LXD().instances
        if instances:
            return instances



import re
from functools import cached_property

from hotsos.core.log import log
from hotsos.core.factory import FactoryBase
from hotsos.core.host_helpers import CLIHelper
from hotsos.core.host_helpers.common import ServiceManagerBase
from hotsos.core.utils import sorted_dict


class PebbleService(object):

    def __init__(self, name, state):
        self.name = name
        self.state = state

    def __repr__(self):
        return "name={}, state={}".format(self.name, self.state)


class PebbleHelper(ServiceManagerBase):
    """ Helper class used to query pebble services. """

    @cached_property
    def services(self):
        """ Return a dict of identified pebble services and their state. """
        _services = {}
        for line in CLIHelper().pebble_services():
            for svc_name_expr in self._service_exprs:
                _expr = r"({})\s+\S+\s+(\S+)\s+.*".format(svc_name_expr)
                ret = re.compile(_expr).match(line)
                if not ret:
                    continue

                name = ret.group(1)
                _services[name] = PebbleService(name=name, state=ret.group(2))

        return _services

    @cached_property
    def processes(self):
        """
        Identify running processes from ps that are associated with resolved
        pebble services.  The search pattern used to identify a service is also
        used to match the process binary/cmd name.

        Accounts for different types of process cmd path e.g.

        /snap/<name>/1830/<svc>
        /usr/bin/<svc>

        and filter e.g.

        /var/lib/<svc> and /var/log/<svc>

        Returns a dictionary of process names along with the number of each.
        """
        _proc_info = {}
        for line in CLIHelper().ps():
            for expr in self._service_exprs:
                cmd = self.get_cmd_from_ps_line(line, expr)
                if cmd:
                    if cmd in _proc_info:
                        _proc_info[cmd] += 1
                    else:
                        _proc_info[cmd] = 1

        return _proc_info

    @property
    def _service_info(self):
        """Return a dictionary of pebble services grouped by state. """
        info = {}
        for svc, obj in sorted_dict(self.services).items():
            state = obj.state
            if state not in info:
                info[state] = []

            info[state].append(svc)

        return info

    @property
    def _process_info(self):
        """Return a list of processes associated with services. """
        return ["{} ({})".format(name, count)
                for name, count in sorted_dict(self.processes).items()]

    @property
    def summary(self):
        """
        Output a dict summary of this class i.e. services, their state and any
        processes run by them.
        """
        return {'pebble': self._service_info,
                'ps': self._process_info}


class ServiceFactory(FactoryBase):
    """
    Factory to dynamically create PebbleService objects for given services.

    Service objects are returned when a getattr() is done on this object using
    the name of the service as the attr name.
    """

    def __getattr__(self, svc):
        log.debug("creating service object for %s", svc)
        return PebbleHelper([svc]).services.get(svc)



from functools import cached_property

# NOTE: we import direct from searchkit rather than hotsos.core.search to
#       avoid circular dependency issues.
from searchkit import (
    FileSearcher,
    SearchDef,
    SequenceSearchDef,
)
from hotsos.core.factory import FactoryBase
from hotsos.core.host_helpers.cli import CLIHelperFile


class AAProfile(object):

    def __init__(self, name):
        self.name = name
        self.mode = ApparmorHelper().get_profile_mode(name)


class ApparmorHelper(object):

    @cached_property
    def profiles(self):
        """
        Fetch all profiles and their mode from apparmor_status.

        @return: dictionary of {<mode>: {'profiles': <list>, 'count': <int>}}
        """
        s = FileSearcher()
        seqdef = SequenceSearchDef(
                     start=SearchDef(r"(\d+) profiles are in (\S+) mode."),
                     body=SearchDef(r"\s+(\S+)"),
                     tag="aastatus")
        info = {}
        with CLIHelperFile() as cli:
            s.add(seqdef, path=cli.apparmor_status())
            results = s.run()
            for section in results.find_sequence_sections(seqdef).values():
                count = 0
                mode = None
                for result in section:
                    if result.tag == seqdef.start_tag:
                        count = int(result.get(1))
                        mode = result.get(2)
                        if mode not in info:
                            info[mode] = {'profiles': [], 'count': count}
                    elif result.tag == seqdef.body_tag:
                        if count == 0:
                            continue

                        info[mode]['profiles'].append(result.get(1))

        return info

    def get_profile_mode(self, name):
        for mode, profiles in self.profiles.items():
            if name in profiles['profiles']:
                return mode

    @property
    def profiles_enforce(self):
        return self.profiles.get('enforce', {}).get('profiles', [])

    @property
    def profiles_complain(self):
        return self.profiles.get('complain', {}).get('profiles', [])

    @property
    def profiles_kill(self):
        return self.profiles.get('kill', {}).get('profiles', [])

    @property
    def profiles_unconfined(self):
        return self.profiles.get('unconfined', {}).get('profiles', [])


class AAProfileFactory(FactoryBase):
    """
    Dynamically create AAProfile objects using profile name.
    """

    def __getattr__(self, profile):
        return AAProfile(profile)



import abc
from functools import cached_property

from hotsos.core.config import HotSOSConfig
from hotsos.core.log import log
from hotsos.core.search import (
    FileSearcher,
    SearchConstraintSearchSince,
)
from hotsos.core.utils import sorted_dict
from hotsos.core.ycheck.engine import (
    YDefsLoader,
    YHandlerBase,
    YDefsSection,
)
from hotsos.core.ycheck.engine.properties.search import CommonTimestampMatcher


CALLBACKS = {}


class EventCallbackNameConflict(Exception):
    pass


class EventCallbackNotFound(Exception):
    pass


class EventCallbackMeta(type):

    def __init__(cls, _name, _mro, members):
        event_group = members.get('event_group')
        if event_group is None:
            return

        for event in members['event_names']:
            event = '{}.{}'.format(event_group, event)
            if event in CALLBACKS:
                msg = "event callback already registered: {}".format(event)
                raise EventCallbackNameConflict(msg)

            CALLBACKS[event] = cls


class EventCheckResult(object):
    """ This is passed to an event check callback when matches are found """

    def __init__(self, defs_section, defs_event, search_results, search_tag,
                 searcher, sequence_def=None):
        """
        @param defs_section: section name from yaml
        @param defs_event: event label/name from yaml
        @param search_results: searchkit.SearchResultsCollection
        @param search_tag: unique tag used to identify the results
        @param searcher: global FileSearcher object
        @param sequence_def: if set the search results are from a
                            searchkit.SequenceSearchDef and are therefore
                            grouped as sections of results rather than a single
                            set of results.
        """
        self.section = defs_section
        self.name = defs_event
        self.search_tag = search_tag
        self.results = search_results
        self.searcher = searcher
        self.sequence_def = sequence_def


class EventProcessingUtils(object):

    @classmethod
    def categorise_events(cls, event, results=None, key_by_date=True,
                          include_time=False, squash_if_none_keys=False,
                          max_results_per_date=None):
        """
        Provides a generic way to categorise events. The default is to group
        events by key which is typically some kind of resource id or event
        label, beneath which the events are grouped by date. If a time value is
        also available, events will be grouped by time beneath their date. It
        may sometimes be useful to group events by date at the top level which
        is also supported here.

        The date, time and key values are extracted from each event search
        result and  are expected to be at indexes 1, 2 and 3 respectively. If
        the search result only has two groups it is assumed that the second is
        "key" i.e. no time is given. If the results argument is provided, this
        will be used instead of event.results.

        @param event: EventCheckResult object
        @param results: optional list of results where each item is a dict with
                        keys 'date', 'key' and optionally 'time'.
        @param key_by_date: by default results are categorised by date but in
                            situations where a small number of event types
                            (keys) are spread across many dates/times it might
                            make sense to categorise by event type.
        @param include_time: If true events will be categorised by time beneath
                             date to provide an extra level of granularity.
        @param squash_if_none_keys: If true and any key is found to be None
                                    (perhaps because a regex pattern did not
                                    match properly) the results for each date
                                    will be squashed to a number/count of
                                    events.
        @param max_results_per_date: an integer such that if key_by_date is
                                     True this will pick the top N entries
                                     with the highest count.
        """
        info = {}
        squash = False
        if results is None:
            # use raw
            results = []
            _fail_count = 0
            for r in event.results:
                # the search expression used much ensure that tese are#
                # available in order for this to work.
                if len(r) > 2:
                    if not set([None, None, None]).intersection([r.get(1),
                                                                 r.get(2),
                                                                 r.get(3)]):
                        results.append({'date': r.get(1), 'time': r.get(2),
                                        'key': r.get(3)})
                    else:
                        _fail_count += 1
                elif len(r) < 1:
                    msg = ("result (tag={}) does not have enough groups "
                           "(min 1) to be categorised - aborting".
                           format(r.tag))
                    raise Exception(msg)
                else:
                    if len(r) < 2:
                        # results with just a date will have None for the key
                        log.debug("result (tag=%s) has just one group which "
                                  "is assumed to be a date and using key=None",
                                  r.tag)

                    results.append({'date': r.get(1), 'key': r.get(2)})

            if _fail_count:
                log.info("event '%s' has %s results with insufficient fields",
                         event.name, _fail_count)

        for r in results:
            if r['key'] is None and squash_if_none_keys:
                squash = True

            ts_time = r.get('time')
            if key_by_date:
                key = r['date']
                value = r['key']
            else:
                key = r['key']
                value = r['date']

            if key not in info:
                info[key] = {}

            if value not in info[key]:
                if ts_time is not None and include_time:
                    info[key][value] = {}
                else:
                    info[key][value] = 0

            if ts_time is not None and include_time:
                if ts_time not in info[key][value]:
                    info[key][value][ts_time] = 1
                else:
                    info[key][value][ts_time] += 1
            else:
                info[key][value] += 1

        if info:
            if squash:
                squashed = {}
                for k, v in info.items():
                    if k not in squashed:
                        squashed[k] = 0

                    for count in v.values():
                        squashed[k] += count

                info = squashed

            # If not using date as key we need to sort the values so that they
            # will have the same order as if they were sorted by date key.
            if not key_by_date:
                for key in info:
                    info[key] = sorted_dict(info[key])
            elif not include_time:
                # sort by value i.e. tally/count
                for key in info:
                    if not isinstance(info[key], dict):
                        break

                    info[key] = sorted_dict(info[key], key=lambda e: e[1],
                                            reverse=True)

            results = sorted_dict(info, reverse=not key_by_date)
            if not (key_by_date and max_results_per_date):
                return results

            shortened = {}
            for date, entries in results.items():
                if len(entries) <= max_results_per_date:
                    shortened[date] = entries
                    continue

                top_n = dict(sorted(entries.items(),
                                    key=lambda e: e[1],
                                    reverse=True)[:max_results_per_date])
                shortened[date] = {'total': len(entries),
                                   "top{}".format(max_results_per_date): top_n}

            return shortened


class EventCallbackBase(EventProcessingUtils, metaclass=EventCallbackMeta):
    # All implementations must set these and event_group must match that used
    # in the handler.
    event_group = None
    event_names = []

    @abc.abstractmethod
    def __call__(self):
        """ Callback method. """


class EventHandlerBase(YHandlerBase, EventProcessingUtils):
    """
    Root name used to identify a group of event definitions. Once all the
    yaml definitions are loaded this defines the level below which events
    for this checker are expected to be found.
    """
    event_group = None

    def __init__(self, *args, searcher=None, **kwargs):
        """
        @param searcher: optional FileSearcher object. When running many event
                         checkers it is more efficient to share a
                         FileSearcher across them so that all searches are
                         done at once.
        """
        super().__init__(*args, **kwargs)
        if not searcher:
            log.debug("creating searcher for event checker")
            searcher = FileSearcher(constraint=SearchConstraintSearchSince(
                                        ts_matcher_cls=CommonTimestampMatcher))

        self._searcher = searcher
        self._event_results = None

    @property
    def searcher(self):
        return self._searcher

    @cached_property
    def event_definitions(self):
        """ Load event definitions from yaml. """
        _event_defs = {}

        plugin = YDefsLoader('events').plugin_defs
        if not plugin:
            return _event_defs

        log.debug("loading defs for subgroup=%s", self.event_group)
        ytree = plugin
        ypath = self.event_group.split('.')
        for i, g in enumerate(ypath):
            if i >= len(ypath) - 1:
                group_defs = ytree.get(g)
            else:
                ytree = ytree.get(g)

        group = YDefsSection(self.event_group, group_defs)
        log.debug("sections=%s, events=%s",
                  len(list(group.branch_sections)),
                  len(list(group.leaf_sections)))

        for event in group.leaf_sections:
            fullname = "{}.{}.{}".format(HotSOSConfig.plugin_name,
                                         event.parent.name, event.name)
            if (HotSOSConfig.event_filter and
                    fullname != HotSOSConfig.event_filter):
                log.info("skipping event %s (filter=%s)", fullname,
                         HotSOSConfig.event_filter)
                continue

            if (not HotSOSConfig.force_mode and event.requires and not
                    event.requires.result):
                log.error("event '%s' pre-requisites not met - "
                          "skipping", event.name)
                return {}

            log.debug("event: %s", event.name)
            log.debug("input: %s (command=%s)", event.input.paths,
                      event.input.command is not None)

            section_name = event.parent.name
            if section_name not in _event_defs:
                _event_defs[section_name] = {}

            for path in event.input.paths:
                if event.input.command:
                    # don't apply constraints to command outputs
                    allow_constraints = False
                else:
                    allow_constraints = True

                event.search.load_searcher(self.searcher, path,
                                           allow_constraints=allow_constraints)

            passthrough = bool(event.search.passthrough_results)
            emeta = {'passthrough': passthrough,
                     'sequence': event.search.sequence_search,
                     'tag': event.search.unique_search_tag}
            _event_defs[section_name][event.name] = emeta

        return _event_defs

    def load(self):
        """ Pre-load event definitions. """
        self.event_definitions

    @property
    def final_event_results(self):
        """ Cache of results in case run() is called again. """
        return self._event_results

    def run(self, results):
        """
        Process each event and call respective callback functions when results
        where found.

        @param results: SearchResultsCollection object.
        """
        if self.final_event_results is not None:
            return self.final_event_results

        if not CALLBACKS:
            raise Exception("need to register at least one callback for "
                            "event handler.")

        log.debug("registered callbacks:\n%s", '\n'.join(CALLBACKS.keys()))
        info = {}
        for section_name, section in self.event_definitions.items():
            for event, event_meta in section.items():
                search_tag = event_meta['tag']
                seq_def = None
                if event_meta['passthrough']:
                    # this is for implementations that have their own means of
                    # retrieving results.
                    search_results = results
                else:
                    seq_def = event_meta['sequence']
                    if seq_def:
                        search_results = results.find_sequence_sections(
                            seq_def)
                        if search_results:
                            search_results = search_results.values()
                    else:
                        search_results = results.find_by_tag(search_tag)

                if not search_results:
                    log.debug("event %s did not yield any results", event)
                    continue

                # We want this to throw an exception if the callback is not
                # defined.
                callback_name = '{}.{}'.format(self.event_group, event)
                if callback_name not in CALLBACKS:
                    msg = ("no callback found for event {}".
                           format(callback_name))
                    raise EventCallbackNotFound(msg)

                callback = CALLBACKS[callback_name]
                event_results_obj = EventCheckResult(section_name, event,
                                                     search_results,
                                                     search_tag,
                                                     self.searcher,
                                                     sequence_def=seq_def)
                log.debug("executing event %s.%s callback '%s'",
                          event_results_obj.section, event,
                          callback_name)
                ret = callback()(event_results_obj)
                if not ret:
                    continue

                # if the return is a tuple it is assumed to be of the form
                # (<output value>, <output key>) where <output key> is used to
                # override the output key for the result which defaults to the
                # event name.
                if isinstance(ret, tuple):
                    out_key = ret[1]
                    ret = ret[0]
                    if not ret:
                        continue
                else:
                    out_key = event

                # Don't clobber results, instead allow them to aggregate.
                if out_key in info:
                    info[out_key].update(ret)
                else:
                    info[out_key] = ret

        if info:
            self._event_results = info
            return info

    def load_and_run(self):
        self.load()
        return self.run(self.searcher.run())



from hotsos.core.plugins.pacemaker import PacemakerChecksBase


class PacemakerSummary(PacemakerChecksBase):
    summary_part_index = 0

    def __0_summary_services(self):
        if self.systemd.services:
            return self.systemd.summary

    def __1_summary_dpkg(self):
        apt = self.apt.all_formatted
        if apt:
            return apt

    def __2_summary_nodes(self):
        nodes = {}
        if self.online_nodes:
            nodes["online"] = self.online_nodes
        if self.offline_nodes:
            nodes["offline"] = self.offline_nodes
        if nodes:
            return nodes



from hotsos.core.log import log
from hotsos.core.host_helpers import PebbleHelper
from hotsos.core.ycheck.engine.properties.requires import (
    intercept_exception,
    ServiceCheckItemsBase,
    YRequirementTypeBase,
)


class PebbleServiceCheckItems(ServiceCheckItemsBase):

    @property
    def _svcs_info(self):
        return PebbleHelper(self._svcs_all)


class YRequirementTypePebble(YRequirementTypeBase):
    """ Provides logic to perform checks on pebble resources. """
    _override_keys = ['pebble']
    _overrride_autoregister = True

    @property
    @intercept_exception
    def _result(self):
        cache_info = {}
        default_op = 'eq'
        _result = True

        items = PebbleServiceCheckItems(self.content)
        if not items.not_installed:
            for svc, settings in items:
                svc_obj = items.installed[svc]
                cache_info[svc] = {'actual': svc_obj.state}
                if settings is None:
                    continue

                processes = None
                if isinstance(settings, str):
                    state = settings
                    ops = [[default_op, state]]
                else:
                    processes = settings.get('processes')
                    if processes:
                        log.debug("checking service processes: %s", processes)

                    op = settings.get('op', default_op)
                    if 'state' in settings:
                        ops = [[op, settings.get('state')]]
                    else:
                        ops = []

                if processes and not items.processes_running(processes):
                    log.debug("one or more processes not running "
                              "- %s", ', '.join(processes))
                    _result = False
                else:
                    cache_info[svc]['ops'] = self.ops_to_str(ops)
                    _result = self.self.apply_ops(ops, input=svc_obj.state)

                if not _result:
                    # bail on first fail
                    break
        else:
            log.debug("one or more services not installed so returning False "
                      "- %s", ', '.join(items.not_installed))
            # bail on first fail i.e. if any not installed
            _result = False

        # this will represent what we have actually checked
        self.cache.set('services', ', '.join(cache_info))
        svcs = ["{}={}".format(svc, settings) for svc, settings in items]
        log.debug('requirement check: %s (result=%s)',
                  ', '.join(svcs), _result)
        return _result



from functools import cached_property

from hotsos.core.log import log
from hotsos.core.utils import sorted_dict
from hotsos.core.search import (
    SearchDef,
    SequenceSearchDef,
    FileSearcher,
)
from hotsos.core.host_helpers import CLIHelperFile


class RabbitMQReport(object):
    """
    Class providing easy access to the contents of a rabbitmqctl report.

    First registers search definitions to execute against rabbitmqctl report
    then runs the search to fetch the information that is then expose through
    properties.

    NOTE: the rabbitmqctl report output differs between versions 3.6.x and
          3.8.x and we try to account for either by providing optional
          regex expressions to match either.
    """

    def __init__(self):
        # save to file so we can search it later
        with CLIHelperFile() as cli:
            searcher = FileSearcher()
            fout = cli.rabbitmqctl_report()
            searcher.add(self.connections_searchdef, fout)
            searcher.add(self.memory_searchdef, fout)
            searcher.add(self.cluster_partition_handling_searchdef, fout)
            searcher.add(self.queues_searchdef, fout)
            self.results = searcher.run()

    @cached_property
    def queues_searchdef(self):
        start = SearchDef([r"^Queues on ([^:]+):",
                           (r"^Listing queues for vhost ([^:]+) "
                            r"...")])
        # NOTE: we don't use a list for the body here because
        # we need to know which expression matched so that we
        # can know in which order to retrieve the columns since
        # their order is inverted between 3.6.x and 3.8.x
        body = SearchDef(r"^(?:<([^.\s]+)[.0-9]+>\s+(\S+)|"
                         r"(\S+)\s+(?:\S+\s+){4}<([^.\s]+)[.0-9]"
                         r"+>)\s+.+")
        end = SearchDef(r"^$")
        return SequenceSearchDef(start=start, body=body, end=end,
                                 tag='queues')

    @cached_property
    def skewed_nodes(self):
        vhosts = self.vhosts
        _skewed_nodes = {}
        skewed_queue_nodes = {}
        global_total_queues = sum(vhost.total_queues for vhost in vhosts)
        for vhost in self.vhosts:
            if not vhost.total_queues:
                continue

            total_pcent = (float(100) / global_total_queues *
                           vhost.total_queues)

            for node, vhost_dist in vhost.node_queue_distributions.items():
                if total_pcent >= 1 and vhost_dist['pcent'] > 75:
                    if node not in skewed_queue_nodes:
                        skewed_queue_nodes[node] = 0

                    skewed_queue_nodes[node] += 1

            # Report the node with the greatest skew of queues/vhost
            if skewed_queue_nodes:
                max_node = None
                for node_name in skewed_queue_nodes:
                    if max_node is None:
                        max_node = node_name
                    elif (skewed_queue_nodes[node_name] >=
                            skewed_queue_nodes[max_node]):
                        max_node = node_name

                if (skewed_queue_nodes[max_node] >
                        _skewed_nodes.get(max_node, 0)):
                    _skewed_nodes[max_node] = skewed_queue_nodes[max_node]

        return _skewed_nodes

    @cached_property
    def vhosts(self):
        seq_def = self.queues_searchdef
        vhosts = []
        for section in self.results.find_sequence_sections(seq_def).values():
            vhost = None
            # ensure we get vhost before the rest
            for result in section:
                if result.tag == seq_def.start_tag:
                    # check both report formats
                    vhost = RabbitMQVhost(result.get(1))
                    break

            for result in section:
                if result.tag == seq_def.body_tag:
                    node_name = result.get(1) or result.get(4)
                    # if we matched the section header, skip
                    if node_name == "pid":
                        continue

                    queue = result.get(2) or result.get(3)
                    # if we matched the section header, skip
                    if queue == "name":
                        continue

                    vhost.node_inc_queue_count(node_name)

            log.debug(vhost.name)
            vhosts.append(vhost)

        return vhosts

    @cached_property
    def connections_searchdef(self):
        start = SearchDef([r"^Connections:$",
                           r"^Listing connections ...$"])
        # Again, the user and protocol columns are inverted
        # between 3.6.x and 3.8.x so we have to catch both and
        # decide.
        body = SearchDef(r"^<(rabbit[^>.]*)(?:[.][0-9]+)+>.+(?:[A-Z]+\s+{[\d,]+}\s+(\S+)|\d+\s+{[\d,]+}\s+\S+\s+(\S+)).+{\"connection_name\",\"([^:]+):\d+:.+$")  # pylint: disable=C0301  # noqa
        end = SearchDef(r"^$")
        return SequenceSearchDef(start=start, body=body, end=end,
                                 tag='connections')

    @cached_property
    def memory_searchdef(self):
        start = SearchDef([r"^Status of node '([^']*)'$",
                           r"^Status of node ([^']*) ...$"])
        body = SearchDef(r"^\s+\[{total,([0-9]+)}.+")
        end = SearchDef(r"^$")
        return SequenceSearchDef(start=start, body=body, end=end,
                                 tag='memory')

    @cached_property
    def cluster_partition_handling_searchdef(self):
        return SearchDef(r"^\s*{cluster_partition_handling,([^}]*)}",
                         tag='cluster_partition_handling')

    @cached_property
    def connections(self):
        _connections = {'host': {}, 'client': {}}
        sd = self.connections_searchdef
        for results in self.results.find_sequence_sections(sd).values():
            for result in results:
                if result.tag == sd.body_tag:
                    host = result.get(1)
                    if host not in _connections['host']:
                        _connections['host'][host] = 1
                    else:
                        _connections['host'][host] += 1

                    # detect 3.6.x or 3.8.x format
                    user = result.get(2)
                    if user is None:
                        user = result.get(3)

                    client_name = result.get(4)
                    if user not in _connections['client']:
                        _connections['client'][user] = {}

                    if client_name not in _connections['client'][user]:
                        _connections['client'][user][client_name] = 1
                    else:
                        _connections['client'][user][client_name] += 1

        if _connections['host']:
            for client, users in _connections['client'].items():
                sorted_users = sorted_dict(users, key=lambda e: e[1],
                                           reverse=True)
                _connections['client'][client] = sorted_users

        return _connections

    @cached_property
    def memory_used(self):
        sd = self.memory_searchdef
        _memory_used = {}
        for results in self.results.find_sequence_sections(sd).values():
            for result in results:
                if result.tag == sd.start_tag:
                    # check both report formats
                    node_name = result.get(1)
                elif result.tag == sd.body_tag:
                    total = result.get(1)
                    mib_used = int(total) / 1024. / 1024.
                    _memory_used[node_name] = "{:.3f}".format(mib_used)

        return _memory_used

    @cached_property
    def partition_handling(self):
        results = self.results.find_by_tag("cluster_partition_handling")
        if not results:
            return

        return results[0].get(1)


class RabbitMQVhost(object):
    def __init__(self, name):
        self.name = name
        self._node_queues = {}

    def node_inc_queue_count(self, node):
        if node not in self._node_queues:
            self._node_queues[node] = 0

        self._node_queues[node] += 1

    @property
    def total_queues(self):
        return sum(self.node_queues.values())

    @property
    def node_queues(self):
        return self._node_queues

    def node_queues_vhost_pcent(self, node):
        return float(100) / self.total_queues * self.node_queues[node]

    @property
    def node_queue_distributions(self):
        dists = {}
        for node, queues in self.node_queues.items():
            if queues:
                vhost_pcent = self.node_queues_vhost_pcent(node)
                dists[node] = {'queues': queues, 'pcent': vhost_pcent}
            else:
                dists[node] = {'queues': 0, 'pcent': 0}

        return dists



import re

from hotsos.core.host_helpers import CLIHelper, HostNetworkingHelper
from hotsos.core.plugins.openstack.common import OpenstackChecksBase
from hotsos.core.plugins.openstack.neutron import (
    IP_HEADER_BYTES,
    GRE_HEADER_BYTES,
    VXLAN_HEADER_BYTES,
)
from hotsos.core.plugins.openvswitch import OpenvSwitchBase
from hotsos.core.issues import (
    IssuesManager,
    OpenstackWarning,
)


class OpenstackNetworkChecks(OpenstackChecksBase):
    summary_part_index = 4

    def __init__(self):
        super().__init__()
        self.cli = CLIHelper()

    @property
    def summary_subkey(self):
        return 'network'

    def _get_port_stat_outliers(self, counters):
        """ For a given port's packet counters, identify outliers i.e. > 1%
        and create a new dict with count and percent values.
        """
        stats = {}
        for rxtx in counters:
            total = sum(counters[rxtx].values())
            for key, value in counters[rxtx].items():
                if key == "packets":
                    continue

                if value:
                    pcent = int(100 / float(total) * float(value))
                    if pcent <= 1:
                        continue

                    if rxtx not in stats:
                        stats[rxtx] = {}

                    stats[rxtx][key] = "{} ({}%)".format(int(value), pcent)

        return stats

    def get_config_info(self):
        config_info = {}
        for project in ['nova', 'neutron', 'octavia']:
            _project = getattr(self, project)
            if _project and _project.bind_interfaces:
                for name, port in _project.bind_interfaces.items():
                    if project not in config_info:
                        config_info[project] = {}

                    config_info[project][name] = port.to_dict()

        return config_info

    def get_phy_port_health_info(self):
        """ Identify ports used by Openstack services, include them in output
        for informational purposes along with their health (dropped packets
        etc) for any outliers detected.
        """
        port_health_info = {}
        for project in ['nova', 'neutron', 'octavia']:
            _project = getattr(self, project)
            if _project and _project.bind_interfaces:
                for port in _project.bind_interfaces.values():
                    if port.stats:
                        stats = self._get_port_stat_outliers(port.stats)
                        if not stats:
                            continue

                        port_health_info[port.name] = stats

        return port_health_info

    def __9_summary_config(self):
        config_info = self.get_config_info()
        if config_info:
            return config_info

    def __10_summary_phy_port_health(self):
        port_health_info = self.get_phy_port_health_info()
        if port_health_info:
            return port_health_info

    def __11_summary_namespaces(self):
        """Populate namespace information dict."""
        ns_info = {}
        for line in self.cli.ip_netns():
            ret = re.compile(r"^([a-z0-9]+)-([0-9a-z\-]+)\s+.+").match(line)
            if ret:
                if ret[1] in ns_info:
                    ns_info[ret[1]] += 1
                else:
                    ns_info[ret[1]] = 1

        if ns_info:
            return ns_info

    def __12_summary_router_port_mtus(self):
        """ Provide a summary of ml2-ovs router port mtus. """
        phy_mtus = set()
        router_mtus = {}
        project = getattr(self, 'neutron')
        if project and project.bind_interfaces:
            for port in project.bind_interfaces.values():
                phy_mtus.add(port.mtu)

            for ns in self.cli.ip_netns():
                # strip index
                ns = ns.partition(" ")[0]
                for nsprefix in ['qrouter', 'snat']:
                    if not ns.startswith(nsprefix):
                        continue

                    for port in HostNetworkingHelper().get_ns_interfaces(ns):
                        for pprefix in ['qr', 'sg']:
                            if not port.name.startswith(pprefix):
                                continue

                            if pprefix not in router_mtus:
                                router_mtus[pprefix] = set()

                            router_mtus[pprefix].add(port.mtu)

            tunnels = OpenvSwitchBase().tunnels
            # ip header
            overhead = IP_HEADER_BYTES
            if 'vxlan' in tunnels:
                overhead += VXLAN_HEADER_BYTES
            else:
                # gre
                overhead += GRE_HEADER_BYTES

            all_router_mtus = set()
            for prefix, mtus in router_mtus.items():
                router_mtus[prefix] = list(mtus)
                for mtu in mtus:
                    all_router_mtus.add(mtu)

            if phy_mtus and all_router_mtus:
                smallest_allowed = min(phy_mtus) - overhead
                if max(all_router_mtus) > smallest_allowed:
                    msg = ("This Neutron L3 agent host has one or more router "
                           "ports with mtu={} which is greater or equal to "
                           "the smallest allowed ({}) on "
                           "the physical network. This will result in dropped "
                           "packets or unexpected fragmentation in overlay "
                           "networks.".format(max(all_router_mtus),
                                              smallest_allowed))
                    IssuesManager().add(OpenstackWarning(msg))

        return router_mtus

    def __13_summary_vm_port_health(self):
        """ For each instance get its ports and check port health, reporting on
        any outliers. """
        if not self.nova.instances:
            return

        port_health_info = {}
        for guest in self.nova.instances.values():
            for port in guest.ports:
                stats = port.stats
                if stats:
                    outliers = self._get_port_stat_outliers(stats)
                    if not outliers:
                        continue

                    if guest.uuid not in port_health_info:
                        port_health_info[guest.uuid] = {}

                    port_health_info[guest.uuid][port.hwaddr] = outliers

        if port_health_info:
            health = {'num-vms-checked': len(self.nova.instances),
                      'stats': port_health_info}
            return health



import abc
import operator
from functools import cached_property

from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.common import YPropertyOverrideBase


def intercept_exception(f):
    """
    If a call raises an AttributeError it will first be handled by
    the propertree engine which can produce confusing/misleading logs
    as to the origin of the exception. This can be used to wrap these
    calls so any exception is printed with its original traceback
    before being re-raised.
    """
    def intercept_exception_inner(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as exc:
            log.error("function '%s' raised exception: %s", f.__name__, exc)
            raise

    return intercept_exception_inner


class CheckItemsBase(abc.ABC):
    """
    Provides a standard way for check items to be defined i.e.

    * a single value e.g. str, int
    * a list
    * a dict

    Items are then accessed by iterating over this object and are always
    presented as a tuple(key, value).
    """

    def __init__(self, raw):
        self._items = {}
        if isinstance(raw, dict):
            self._items = raw
        elif isinstance(raw, list):
            self._items = {i: None for i in raw}
        else:
            self._items = {raw: None}

    def __iter__(self):
        for item in self._items.items():
            yield item


class PackageCheckItemsBase(CheckItemsBase):

    @cached_property
    def packages_to_check(self):
        return [item[0] for item in self]

    @property
    @abc.abstractmethod
    def packaging_helper(self):
        """
        Returns and implementation of host_helpers.packaging.PackageHelperBase
        e.g. APTPackagehelper or SnapPackageHelper.
        """

    @cached_property
    def installed(self):
        _installed = []
        for p in self.packages_to_check:
            if self.packaging_helper.is_installed(p):
                _installed.append(p)

        return _installed

    @cached_property
    def not_installed(self):
        return set(self.packages_to_check).difference(self.installed)


class OpsUtils(object):

    def ops_to_str(self, ops):
        """
        Convert an ops list of tuples to a string. This is typically used when
        printing in a msg or storing in the cache.
        """
        if not ops:
            return ""

        _result = []
        for op in ops:
            item = str(op[0])
            if len(op) > 1:
                if isinstance(op[1], str):
                    item = "{} \"{}\"".format(item, op[1])
                else:
                    item = "{} {}".format(item, op[1])

            _result.append(item)

        return ' -> '.join(_result)

    def _apply_op(self, op, opinput=None, expected=None, force_expected=False):
        """
        @param expected: can be a value or variable name that needs to be
                         resolved. Variable names are identified by having a
                         "$" prefix.
        """
        log.debug("op=%s, input=%s, expected=%s, force_expected=%s", op,
                  opinput, expected, force_expected)
        try:
            if expected is not None or force_expected:
                return getattr(operator, op)(opinput, expected)

            return getattr(operator, op)(opinput)
        except Exception:
            log.error("failed to apply operator '%s'", op)
            raise

    def apply_ops(self, ops, opinput=None, normalise_value_types=False):
        """
        Takes a list of operations and processes each one where each takes as
        input the output of the previous.

        @param ops: list of tuples of operations and optional args.
        @param opinput: the value that is used as input to the first operation.
        @param normalise_value_types: if an operation has an expected value and
                                      and this is True, the type of the input
                                      will be cast to that of the expectced
                                      value.
        """
        log.debug("ops=%s, input=%s", ops, opinput)
        if not isinstance(ops, list):
            raise Exception("Expected list of ops but got {}".
                            format(ops))

        for op in ops:
            expected = None
            force_expected = False
            if len(op) > 1:
                # if an expected value was provided we must use it regardless
                # of what it is.
                force_expected = True
                expected = op[1]

                if expected is not None:
                    if isinstance(expected, str) and expected.startswith("$"):
                        varname = expected.partition("$")[2]
                        varval = self.context.vars.resolve(varname)
                        expected = varval

                    if normalise_value_types:
                        log.debug("normalising type(input)=%s to "
                                  "type(expected)=%s", type(opinput),
                                  type(expected))
                        opinput = type(expected)(opinput)

            opinput = self._apply_op(op[0], opinput=opinput, expected=expected,
                                     force_expected=force_expected)

        return bool(opinput)


class YRequirementTypeBase(YPropertyOverrideBase, OpsUtils):
    _overrride_autoregister = False

    @property
    @abc.abstractmethod
    def _result(self):
        """ Assert whether the requirement is met.

        Returns True if met otherwise False.
        """

    @property
    def result(self):
        try:
            return self._result
        except Exception:
            # display traceback here before it gets swallowed up.
            log.error("requires.%s.result raised the following",
                      self.__class__.__name__)
            raise


class YRequirementTypeWithOpsBase(YRequirementTypeBase):

    @property
    def default_ops(self):
        return [['truth']]

    @property
    def ops(self):
        if not isinstance(self.content, dict):
            return self.default_ops

        return self.content.get('ops', self.default_ops)


class ServiceCheckItemsBase(CheckItemsBase):

    @cached_property
    def _started_after_services(self):
        svcs = []
        for _, settings in self:
            if not isinstance(settings, dict):
                continue

            svc = settings.get('started-after')
            if svc:
                svcs.append(svc)

        return svcs

    @cached_property
    def _services_to_check(self):
        return [item[0] for item in self]

    @property
    def _svcs_all(self):
        """
        We include started-after services since if a check has specified one it
        is expected to exist for that check to pass.
        """
        return self._services_to_check + self._started_after_services

    @property
    @abc.abstractmethod
    def _svcs_info(self):
        """ ServiceManagerBase implementation with _svcs_all as input. """

    @cached_property
    def not_installed(self):
        return set(self._svcs_all).difference(self.installed)

    @cached_property
    def installed(self):
        return self._svcs_info.services

    def processes_running(self, processes):
        """ Check any processes provided. """
        a = set(processes)
        b = set(self._svcs_info.processes.keys())
        return a.issubset(b)
